# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the "Software"), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

class CachedScoreValue:
    def __init__(self, value, score, timestamp):
        self.value = value
        self.score = score
        self.timestamp = timestamp


class HeapWithInverseIndex():
    """
    Heap where elements can be retrieved using a "key". This allows e.g. for
    updating arbitrary elements. It also supports invalidating arbitrary
    elements in a (time) efficient way.

    The key can be customized through the key_function parameter.
    """
    def __init__(self,
                 ini_vector=None,
                 compare_function=lambda a, b: a > b,
                 value_score_function=lambda x: x,
                 key_function=lambda x: x,
                 use_score_caching=False):
        self.cmp = lambda a, b: compare_function(self._get_value_score(a), self._get_value_score(b))
        self.timestep = 0
        self.use_score_caching = use_score_caching
        self.value_score_function = value_score_function
        self.inverse_index = {}
        self.keyf = lambda v: key_function(v.value)
        self.invalid_pos = set()
        self.data = []
        if ini_vector:
            for (i, x) in enumerate(ini_vector):
                value = CachedScoreValue(x, None, None)
                self.data.append(value)
                self.inverse_index[self.keyf(value)] = i
            self.heapify()

    def _get_value_score(self, cached_score_data):
        if not self.use_score_caching:
            return self.value_score_function(cached_score_data.value)
        else:
            if cached_score_data.timestamp != self.timestep:
                cached_score_data.score = self.value_score_function(cached_score_data.value)
                cached_score_data.timestamp = self.timestep
            return cached_score_data.score

    def increase_timestep(self):
        self.timestep += 1

    def __len__(self):
        return len(self.data)

    def get_max(self):
        return self.data[0].value

    def _swap_pos(self, i, j):
        ori_i_is_invalid = i in self.invalid_pos
        ori_j_is_invalid = j in self.invalid_pos
        (self.data[i], self.data[j]) = (self.data[j], self.data[i])

        # Note: we have swapped positions now
        if ori_i_is_invalid:
            self.invalid_pos.add(j)
        else:
            self.invalid_pos.discard(j)
            self.inverse_index[self.keyf(self.data[j])] = j

        if ori_j_is_invalid:
            self.invalid_pos.add(i)
        else:
            self.invalid_pos.discard(i)
            self.inverse_index[self.keyf(self.data[i])] = i

    def insert(self, elem):
        if self.invalid_pos:
            # Take one of the invalidated positions
            target_pos = self.invalid_pos.pop()
            new_elem = CachedScoreValue(elem, None, None)
            self.data[target_pos] = new_elem
            self.inverse_index[self.keyf(new_elem)] = target_pos
            self.sift_down(target_pos)
            self.sift_up(target_pos)
        else:
            new_elem = CachedScoreValue(elem, None, None)
            self.inverse_index[self.keyf(new_elem)] = len(self.data)
            self.data.append(new_elem)
            self.sift_up(len(self.data) - 1)

    def get(self, key):
        pos = self.inverse_index.get(key)
        if pos is not None:
            return self.data[pos].value
        else:
            return None

    def pop_max_cached_value(self):
        if not self.data:
            raise IndexError("Heap is empty")
        if 0 in self.invalid_pos:
            self.push_invalid(0)
        if 0 in self.invalid_pos:
            raise IndexError("No valid elements in heap")

        maximum = self.data[0]
        del self.inverse_index[self.keyf(self.data[0])]
        if len(self.data) > 1:
            self.clean_invalid()
            last_element_pos = len(self.data) - 1
            if last_element_pos in self.invalid_pos:
                # TODO: This shouldn't happen...
                self.invalid_pos.add(0)
                self.invalid_pos.discard(last_element_pos)
            self.data[0] = self.data.pop()
            self.inverse_index[self.keyf(self.data[0])] = 0
            self.sift_down(0)
        else:
            self.data = []
            self.invalid_pos = set()
        #~ print("%.2f%%" % (100 * float(len(self.invalid_pos)) / len(self.data)))
        return maximum

    def pop_max(self):
        return self.pop_max_cached_value().value

    def parent(self, i):
        return int((i - 1) / 2)

    def left_child(self, i):
        return 2 * i + 1

    def right_child(self, i):
        return 2 * i + 2

    def has_child(self, i):
        return self.left_child(i) < len(self.data)

    def heapify(self):
        start = self.parent(len(self.data) - 1)
        while start >= 0:
            self.sift_down(start)
            start -= 1

    def substitute_down_key(self, key, new_value):
        pos = self.inverse_index[key]
        self.data[pos].value = new_value
        self.data[pos].timestamp = None
        self.sift_down(pos)

    def substitute_up_key(self, key, new_value):
        pos = self.inverse_index[key]
        self.data[pos].value = new_value
        self.data[pos].timestamp = None
        self.sift_up(pos)

    def remove_key(self, key):
        pos = self.inverse_index.get(key)
        if pos is not None:
            self.remove(pos)
            del self.inverse_index[key]

    def invalidate_key(self, key):
        pos = self.inverse_index.get(key)
        if pos is not None:
            self.invalid_pos.add(pos)
            del self.inverse_index[key]

    def push_invalid(self, pos):
        root = pos
        if pos not in self.invalid_pos:
            return
        if self.has_child(root):
            # Push an invalid node towards the bottom of the heap
            left_child = self.left_child(root)
            right_child = self.right_child(root)

            if left_child in self.invalid_pos:
                self.push_invalid(left_child)
            if right_child < len(self.data):
                if right_child in self.invalid_pos:
                    self.push_invalid(right_child)

                if right_child not in self.invalid_pos and left_child not in self.invalid_pos:
                    # Both are valid
                    exchange_with = left_child if self.cmp(self.data[left_child], self.data[right_child]) else right_child
                elif left_child not in self.invalid_pos:
                    # Left is the only valid
                    exchange_with = left_child
                else:
                    # Either both are invalid or right is the only valid
                    exchange_with = right_child
            else:
                exchange_with = left_child

            if exchange_with not in self.invalid_pos:
                self._swap_pos(root, exchange_with)
                self.push_invalid(exchange_with)
                self.clean_invalid()

    def clean_invalid(self):
        removed = 0
        while (len(self.data) - 1) in self.invalid_pos:
            self.invalid_pos.discard(len(self.data) - 1)
            self.data.pop()
            removed += 1

    def push_all_invalid(self):
        for pos in range(len(self.data)-1, -1, -1):
            if pos in self.invalid_pos:
                self.push_invalid(pos)

    def sift_down(self, pos):
        root = pos
        if root in self.invalid_pos:
            return
        while self.has_child(root):
            left_child = self.left_child(root)
            right_child = self.right_child(root)

            if left_child in self.invalid_pos:
                self.push_invalid(left_child)
            if right_child < len(self.data) and right_child in self.invalid_pos:
                self.push_invalid(right_child)

            if left_child in self.invalid_pos:
                if right_child >= len(self.data) or right_child in self.invalid_pos:
                    break
                else:
                    child = right_child
            else:
                if right_child >= len(self.data) or right_child in self.invalid_pos:
                    child = left_child
                else:
                    child = left_child if self.cmp(self.data[left_child], self.data[right_child]) else right_child

            if self.cmp(self.data[child], self.data[root]):
                self._swap_pos(root, child)
                root = child
            else:
                break
        self.clean_invalid()

    def sift_up(self, pos):
        root = pos
        if root in self.invalid_pos:
            return
        while root > 0:
            parent = self.parent(root)
            if parent in self.invalid_pos:
                self._swap_pos(root, parent)
                # Because the root was invalid, we have to consider if the
                # other subtree may have a bigger element
                self.sift_down(parent)
                root = parent
            elif self.cmp(self.data[root], self.data[parent]):
                self._swap_pos(root, parent)
                root = parent
            else:
                break
        self.clean_invalid()

    def remove(self, pos):
        if pos != len(self.data) - 1:
            self._swap_pos(pos, len(self.data) - 1)
            self.invalid_pos.discard(len(self.data) - 1)
            self.data.pop()
            if pos > 0 and self.cmp(self.data[pos], self.data[self.parent(pos)]):
                self.sift_up(pos)
            else:
                self.sift_down(pos)
        else:
            self.invalid_pos.discard(len(self.data) - 1)
            self.data.pop()

    def update_compare_function(self, new_cmp):
        self.cmp = new_cmp
        self.heapify()

    def dfs(self, pos):
        stack = [pos]
        while stack:
            root = stack.pop()
            yield root
            child = self.left_child(root)
            if child < len(self.data):
                stack.append(child)
                child += 1  # Right child
                if child < len(self.data):
                    stack.append(child)

    def check_heap_property(self, root):
        if root in self.invalid_pos:
            return True # Actually we should check the children, but we will call it in the whole heap
        holds = True
        for node in self.dfs(root):
            if node == root:
                continue
            holds = node in self.invalid_pos or self.cmp(self.data[root], self.data[node])
            if not holds:
                print("Does not hold  root(%d) = %d node(%d) = %d" % (root, self.data[root], node, self.data[node]))
                break
        return holds



"""
Goal: Convert a HuggingFace byte-based vocabulary, which is stored in text form with every byte represented by one of 256
      alphabet characters, back to Unicode. This is hard because the alphabet they give is NOT sorted, so you need to
      discover the mapping between alphabet characters and byte values yourself.

      All of this work is kinda moot given the bytes_to_unicode mapping in HuggingFace's GPT2Tokenizer source code.
"""
from tokenizers import pre_tokenizers

BYTE_ALPHABET = pre_tokenizers.ByteLevel.alphabet()
BYTE_ALPHABET.sort()


def huggingFaceIndexToByte(i: int):
    """
    Replicates the index-to-byte mapping found by the comparative graph made elsewhere.
    A couple of inputs were interpolated because they didn't appear on the graph:
        124, 125
        177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187
    """
    if i < 0 or i > 255:
        raise ValueError("Index must be in the interval [0, 255].")
    elif i < 94:
        return i+33
    elif i < 106:
        return i+67
    elif i < 188:
        return i+68
    elif i < 221:
        return i-188
    elif i < 255:
        return i-94
    else:  # i == 255 is special
        return i-82


def decodeHuggingFaceBytes(pseudochar_string: str) -> str:
    """
    This is effectively a re-implementation of convert_tokens_to_string but without needing any particular tokeniser object.
    Note that when a half-codepoint is on the edge of a token, it is impossible to translate it to text.

    In fact, this has other limitations: if you translate Ã« to ë but you don't translate Ã or « to anything else, then
    merges can no longer be concatenated to get the resulting type: aaÃ + «bb (6 chars) results in aaëbb (5 chars).

    This is an issue, because it is an assumption in my BPE code: the type that results from a merge is found via a call
    to "".join(parts). If you do this in the above example, you get aaÃ«bb. This is added to the vocab. Then you find
    subsequent merges that contain the part aaëbb, which 1. is not known to the tokeniser vocab, and 2. even if it was,
    would not have any incoming merge. The history with its past parts is lost.

    There is a way to solve it: apply decodeHuggingFaceBytes() in real-time rather than as a preprocessing step for your
    vocab file. I.e.: apply it to your types, your parts, and to the combination of parts, before entering them into the
    graph. You do have to make sure to never apply this function twice to the same string, because then you can get a
    double collapse (some bytes -> Ã and some bytes -> « and then Ã« -> ë).

    Another limitation is obviously that if a language actually uses the pseudo-characters as actual characters, you are
    going to misinterpret those and possibly destroy the language. (This is only applicable if the vocab is NOT byte-level.)
    """
    all_bytes = bytes([huggingFaceIndexToByte(BYTE_ALPHABET.index(c)) for c in pseudochar_string])
    try:
        return all_bytes.decode("utf-8")
    except:
        trim = 1
        while trim < len(all_bytes):
            try:
                return pseudochar_string[:trim] + all_bytes[trim:].decode("utf-8")
            except:
                try:
                    return all_bytes[:trim].decode("utf-8") + pseudochar_string[trim:]
                except:
                    trim += 1

        return pseudochar_string  # Trimming all bytes is equivalent to just returning the garbage you received.


def simplifiedByteMapper(s: str):
    """
    Based on observations, merges in e.g. Dutch only really need accents to be converted.
    Obviously, having been trained on internet data, RobBERT also has Chinese, Russian, etc...
    yet for our purposes, those merges will never be touched and hence they are useless.

    A speed test shows that this method is more expensive than HuggingFace's native decoder,
    except when the input contains spaces that need to be protected by splitting and joining.
    """
    return s \
        .replace("Ã¤", "ä").replace("Ã¡", "á").replace("Ãł", "à") \
        .replace("ÃĦ", "Ä").replace("Ãģ", "Á").replace("ÃĢ", "À") \
        .replace("Ã«", "ë").replace("Ã©", "é").replace("Ã¨", "è").replace("Ãª", "ê") \
        .replace("Ãĭ", "Ë").replace("Ãī", "É").replace("ÃĪ", "È") \
        .replace("Ã¯", "ï").replace("ÃŃ", "í").replace("Ã¬", "ì") \
        .replace("Ãı", "Ï").replace("Ãį", "Í").replace("ÃĮ", "Ì") \
        .replace("Ã¶", "ö").replace("Ã³", "ó").replace("Ã²", "ò") \
        .replace("Ãĸ", "Ö").replace("Ãĵ", "Ó").replace("ÃĴ", "Ò") \
        .replace("Ã¼", "ü").replace("Ãº", "ú").replace("Ã¹", "ù") \
        .replace("Ãľ", "Ü").replace("Ãļ", "Ú").replace("ÃĻ", "Ù") \
        .replace("âĤ¬", "€") \
        .replace("ÃŁ", "ß")



from typing import Dict
from pathlib import Path

import json
from tokenizers import Tokenizer, models, normalizers, pre_tokenizers, trainers, decoders

from bpe_knockout.util.storage import SennrichTokeniserPath, HuggingFaceTokeniserPath
from ..datahandlers.wordfiles import wordfileToBpeCorpus
from bpe_knockout._lib.sbpe.learn_bpe import learn_bpe

from transformers import SpecialTokensMixin
PAD = "<pad>"
BOS = "<s>"
EOS = "</s>"
MSK = "<mask>"
UNK = "<unk>"
SPECIAL_TYPES = SpecialTokensMixin(
    pad_token=PAD,
    bos_token=BOS,
    eos_token=EOS,
    mask_token=MSK,
    unk_token=UNK
)  # The above argument mapping is reconstructed with .special_tokens_map; the list of values is .all_special_tokens

SOW = "Ġ"


class BPETrainer:

    def __init__(self, vocab_size: int, byte_based: bool):
        self.size   = vocab_size
        self.soweow = SowEowSpecification(detached=True, start_not_end=True, character=SOW)
        self.byte_based = byte_based
        self.normaliser = normalizers.NFKC()

    def train(self, wordfile: Path, out_folder: Path):
        paths = SennrichTokeniserPath(folder=out_folder)
        path_vocab, path_merges = paths.getPaths()

        # Learn merges
        if self.byte_based:
            pretokeniser = pre_tokenizers.ByteLevel(add_prefix_space=False)  # Also a punctuation tokeniser!
            preprocessor = lambda word: [token for token, _ in pretokeniser.pre_tokenize_str(self.normaliser.normalize_str(word))]
        else:
            preprocessor = lambda word: [self.normaliser.normalize_str(word)]

        with open(path_merges, "w", encoding="utf-8") as out_handle:
            with open(wordfile, "r", encoding="utf-8") as in_handle:
                learn_bpe([in_handle], out_handle, num_symbols_ori=self.size, total_symbols=True,
                          is_dict=True, preprocessor=preprocessor, marker=self.soweow)

        # Deduce vocab
        vocab = BPETrainer.deduceVocabFromMerges(path_merges, byte_based=self.byte_based)
        with open(path_vocab, "w", encoding="utf-8") as out_handle:
            json.dump(vocab, out_handle, ensure_ascii=False, indent=4)

    @staticmethod
    def deduceVocabFromMerges(mergefile: Path, byte_based: bool) -> Dict[str, int]:
        # Summarise merges
        with open(mergefile, "r", encoding="utf-8") as in_handle:
            merges = [line.strip() for line in in_handle if line != "#version: 0.2\n"]

        used_types     = set()
        produced_types = set()
        for merge in merges:
            parts = merge.split()
            used_types.update(parts)
            produced_types.add("".join(parts))

        # Get alphabet
        alphabet = pre_tokenizers.ByteLevel().alphabet() if byte_based else used_types - produced_types

        # Combine everything
        vocab = {c: i for i, c in enumerate(
            SPECIAL_TYPES.all_special_tokens +
            sorted(alphabet) +
            list(produced_types)
        )}

        return vocab

    def train_hf(self, wordfile: Path, out_folder: Path):
        """
        HuggingFace equivalent. For German: starts out extremely slow
        (giving an ETA of 500 000 hours), but finishes in under 2 hours.
        """
        # Model: no normaliser (because RobBERT doesn't have one) and no decoder (because training is back-end-only).
        tokeniser = Tokenizer(models.BPE())
        if self.byte_based:
            tokeniser.pre_tokenizer = pre_tokenizers.ByteLevel(add_prefix_space=False, trim_offsets=True)
            tokeniser.decoder       = decoders.ByteLevel()
            # TODO: Spaces are a massive issue.
            #     - Somewhere in the HuggingFace library, out of the user's control, all spaces passed to the training loop are
            #       converted to Ġ. This means you can't add it yourself (its UTF-8 encoding is \xc4\xa0 which is represented as Äł
            #       in the tokeniser file -- because \xc4 is Ä in Latin-1 and \xa0 is somehow remapped to ł by HuggingFace, given that
            #       the only charset where ł has 1 byte is latin2 where it is actually \xb3).
            #     - So then, how do you ensure intra-word boundaries (around hyphens)? Can't use a space...
            #       > ByteLevel pretokenisation takes care of this by accident, but this doesn't work if you want an end-of-word.
        else:
            tokeniser.pre_tokenizer = pre_tokenizers.Metaspace(replacement=SOW)
            tokeniser.decoder       = decoders.Metaspace(replacement=SOW)

        # Trainer interface according to https://huggingface.co/docs/tokenizers/api/trainers (ignore the type hints that complain):
        trainer = trainers.BpeTrainer(
            vocab_size=self.size,
            show_progress=True,
            special_tokens=SPECIAL_TYPES.all_special_tokens,
            initial_alphabet=tokeniser.pre_tokenizer.alphabet() if self.byte_based else []  # after https://huggingface.co/docs/tokenizers/training_from_memory
        )
        tokeniser.train_from_iterator(wordfileToBpeCorpus(wordfile, do_pretokenise=False), trainer=trainer)

        # Save
        save_path = out_folder / f"BPE_from_{wordfile.stem}.json"
        hf = HuggingFaceTokeniserPath(json_path=save_path)  # Calls .mkdir, which is important because otherwise the next line fails.
        tokeniser.save(path=save_path.as_posix())

        # Turn into vocab.json + merges.txt
        vocab, merges = SennrichTokeniserPath(folder=out_folder).getPaths()
        with open(vocab, "w", encoding="utf-8") as out_handle:
            json.dump(hf.loadVocabulary(), out_handle, ensure_ascii=False, indent=4)
        with open(merges, "w", encoding="utf-8") as out_handle:
            out_handle.writelines([merge + "\n" for merge in hf.loadMerges()])



import re

from string import punctuation
punctuation = punctuation + "€£…‘’“”„«»–"  # Add some European punctuations.
punctuation = punctuation.replace("\\", "") + "\\"  # Put backslash in the back. Makes the pattern clearer.
punctuation = "-" + punctuation.replace("-", "")    # Put hyphen in the front. Prevents regex from thinking it's a span.
punctuation_regex_str = "[" + punctuation.replace("\\", "\\\\").replace("[", "\\[").replace("]", "\\]").replace("-", "\\-") + "]+"

###
all_allowed_characters = re.compile(r"([A-Za-z0-9]|[ÁáÉéÍíÓóÚúÄäËëÏïÖöÜüÀàÈèÌìÒòÙùÂâÊêÎîÔôÛû]|°[CFK]|[" + punctuation_regex_str[1:-2] + r"°®©•·¿±×÷⁄≠≤≥‰∞])")

dashes_to_replace = re.compile(r"[‐−—]")
FILTERED_WHITESPACE = re.compile(r"[​﻿]")
CONTROL_CHARACTERS = re.compile(r"[‎‏­\u202C]")  # ­ is a "soft hyphen" which isn't whitespace, but its entire purpose is actually to NOT to be a hyphen if you have the space.
dashes_to_replace = re.compile(r"[‐−—]")
###

MORE_EUROPEAN_ACCENTS = re.compile(r"[ÃãāÅåấầẩắạậėệếēẹÕõŌōØøœỚộỗőốổồờợịĪīįűūưųựứÇçÑñƒß]")
TURKISH_SPECIALS   = re.compile(r"[ŞşĞğİı]")
ICELANDIC_SPECIALS = re.compile(r"[ÐðÞþÆæý]")
ROMANIAN_SPECIALS  = re.compile(r"[ĂăȘșȚțŢţ]")
POLISH_SPECIALS    = re.compile(r"[ĄąĆćĘęŁłŃńÓóŚśŹźŻż]")
CZECH_SPECIALS     = re.compile(r"[ÁáČčĎďÉéĚěÍíŇňÓóŘřŠšŤťÚúŮůÝýŽž]")

CYRILLIC_CHARACTERS = re.compile(r"[\u0400-\u04FF]")
GREEK_CHARACTERS   = re.compile(r"[\u0370-\u03FF]")  # https://en.wikipedia.org/wiki/Greek_script_in_Unicode
GEORGIAN_CHARACTERS = re.compile(r"[\u10A0-\u10FF]")
SYRO_ARAMAIC_CHARACTERS = re.compile(r"[\u0700-\u074F]")
GURMUKHI_CHARACTERS     = re.compile(r"[\u0A00-\u0A7F]")
CHINESE_CHARACTERS = re.compile(r"[\u4e00-\u9fff]")  # https://stackoverflow.com/a/34587623/9352077
INDIAN_CHARACTERS  = re.compile(r"[\u0900-\u097F]|[\u0980-\u09FF]|[\u0A80-\u0AFF]|[\u0D00-\u0D7F]")  # https://en.wikipedia.org/wiki/Brahmic_scripts#Unicode_of_Brahmic_scripts
SRI_LANKAN_CHARACTERS = re.compile(r"[\u0D80-\u0DFF]")
LAO_CHARACTERS     = re.compile(r"[\u0E80-\u0EFF]")
THAI_CHARACTERS    = re.compile(r"[\u0E00-\u0E7F]")  # https://en.wikipedia.org/wiki/Thai_script
JAPANESE_CHARACTERS = re.compile(r"[\u4E00-\u9FBF]|[\u3040-\u309F]|[\u30A0-\u30FF]")  # https://en.wikipedia.org/wiki/Japanese_writing_system
KOREAN_CHARACTERS  = re.compile(r"[\u1100-\u11FF]|[\u3130-\u318F]|[\uA960-\uA97F]|[\uAC00-\uD7AF]|[\uD7B0-\uD7FF]")  # https://en.wikipedia.org/wiki/Hangul
HEBREW_CHARACTERS  = re.compile(r"[\u0590-\u05FF]")  # https://en.wikipedia.org/wiki/Hebrew_alphabet
ARABIC_CHARACTERS  = re.compile(r"[\u0600-\u06FF]|[\uFB50-\uFDFF]")  # https://en.wikipedia.org/wiki/Arabic_script; the second range is gigantic characters like (copy-paste in a browser) ﷽

WEIRD_ACCENTS    = re.compile("[" + r"́" + r"̃" + r"ʿʾ" + r"‚ʼ′ꞌ̍" + "]")
WEIRD_CHARACTERS = re.compile(r"[¤¢†‡¬¶§↓↑←→¡‹›！]")
UNICODE_REPLACEMENT_CHAR = re.compile(r"\uFFFD")
UNICODE_PRIVATE_USE = re.compile(r"[\uE000-\uF8FF]")  # https://www.compart.com/en/unicode/block/U+E000
EMOJI              = re.compile(r"[\u25a0-\u27bf]|\ud83c[\ud000-\udfff]|\ud83d[\ud000-\udfff]|\ud83e[\ud000-\udfff]")  # https://stackoverflow.com/a/67705964/9352077
C_AND_R            = re.compile(r"[\u00a9\u00ae]")
# BROKEN_ACCENTS = re.compile(r"���")  # ëïé

UNICODE_EXPLANATIONS = {
    "Chinese": CHINESE_CHARACTERS,
    "Indic": INDIAN_CHARACTERS,
    "Sri-Lankan": SRI_LANKAN_CHARACTERS,
    "Lao": LAO_CHARACTERS,
    "Thai": THAI_CHARACTERS,
    "Japanese": JAPANESE_CHARACTERS,
    "Korean": KOREAN_CHARACTERS,
    "Hebrew": HEBREW_CHARACTERS,
    "Arabic": ARABIC_CHARACTERS,
    "Greek": GREEK_CHARACTERS,
    "Cyrillic": CYRILLIC_CHARACTERS,
    "Turkish": TURKISH_SPECIALS,
    "Syro-Aramaic": SYRO_ARAMAIC_CHARACTERS,
    "Polish": POLISH_SPECIALS,
    "Czech": CZECH_SPECIALS,
    "Romanian": ROMANIAN_SPECIALS,
    "Gurmukhi": GURMUKHI_CHARACTERS,
    "Georgian": GEORGIAN_CHARACTERS,
    "Icelandic": ICELANDIC_SPECIALS,

    "European special": MORE_EUROPEAN_ACCENTS,

    "Weird accents": WEIRD_ACCENTS,
    "Weird symbols": WEIRD_CHARACTERS,
    "Emoji": EMOJI,
    "Unicode repl": UNICODE_REPLACEMENT_CHAR,
    "Unicode private": UNICODE_PRIVATE_USE
}
